/*
  Copyright (C) MINZKN.COM
  All rights reserved.
  Code by JaeHyuk Cho <mailto:minzkn@minzkn.com>
*/

#if !defined(_ISOC99_SOURCE)
# define _ISOC99_SOURCE (1L)
#endif

#if !defined(_GNU_SOURCE)
# define _GNU_SOURCE (1L)
#endif

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <pthread.h>

#include <sqlite3.h>

#define def_test_sqlite3_database_filename "test_sqlite3.db"

#define def_test_sqlite3_process_count (4)
#define def_test_sqlite3_thread_count (4)

#define def_test_sqlite3_drop_schema "DROP TABLE IF EXISTS test_sqlite3"
#define def_test_sqlite3_schema "CREATE TABLE IF NOT EXISTS test_sqlite3 (value1 VARCHAR(32) NOT NULL DEFAULT '' PRIMARY KEY, value2 VARCHAR(32) NOT NULL DEFAULT '')" 

#define def_test_sqlite3_insert1 "INSERT INTO test_sqlite3 (value1,value2) VALUES ('test1', '')"
#define def_test_sqlite3_insert2 "INSERT INTO test_sqlite3 (value1,value2) VALUES ('test2', '')"
#define def_test_sqlite3_insert3 "INSERT INTO test_sqlite3 (value1,value2) VALUES ('test3', '')"

static int test_sqlite3_busy_handler(void *s_argument, int s_count);
static int test_sqlite3_query(sqlite3 *s_sqlite3, pid_t s_pid, int s_id, const char *s_query);
static void *test_sqlite3_thread(void *s_argument);
static int test_sqlite3_callback(void *s_user_context, int s_argc, char **s_argv, char **s_field_name);

int main(int s_argc, char **s_argv);

static int test_sqlite3_busy_handler(void *s_argument, int s_count)
{
    (void)s_argument;
    (void)s_count;

    usleep(1);

    return(1);
}

static int test_sqlite3_query(sqlite3 *s_sqlite3, pid_t s_pid, int s_id, const char *s_query)
{
    int s_retry;
    int s_sqlite3_check;
    char *s_error_message;

    s_retry = 0;
    do {
        /* SQLITE_API int sqlite3_exec(sqlite3*, const char *sql, int (*callback)(void*,int,char**,char**), void *, char **errmsg); */
        s_sqlite3_check = sqlite3_exec(s_sqlite3, s_query, test_sqlite3_callback, (void *)0, (char **)(&s_error_message));
        switch(s_sqlite3_check) {
            case SQLITE_OK:
                s_retry = 0;
                break;
            case SQLITE_LOCKED:
                ++s_retry;
                (void)fprintf(stderr, "sqlite3_exec locked (pid=%d, id=%d, retry=%d) : error_code=%d, error_message=\"%s\"\n", (int)s_pid, s_id, s_retry, s_sqlite3_check, s_error_message);
                break;
            case SQLITE_BUSY:
                ++s_retry;
                (void)fprintf(stderr, "sqlite3_exec busy (pid=%d, id=%d, retry=%d) : error_code=%d, error_message=\"%s\"\n", (int)s_pid, s_id, s_retry, s_sqlite3_check, s_error_message);
                break;
            default:
                ++s_retry;
                (void)fprintf(stderr, "sqlite3_exec failed (pid=%d, id=%d, retry=%d) : error_code=%d, error_message=\"%s\"\n", (int)s_pid, s_id, s_retry, s_sqlite3_check, s_error_message);
                break;
        }
    }while(s_retry > 0);

    return(s_sqlite3_check);
}

static void *test_sqlite3_thread(void *s_argument)
{
    pid_t s_pid = getpid();
    int s_id = (s_argument == ((void *)0)) ? (-1) : (*((int *)s_argument));
    int s_sqlite3_check;

    sqlite3 *s_sqlite3 = (sqlite3 *)0;

    char s_query[ 1 << 10 ];

    (void)fprintf(stdout, "running_thread (pid=%d, id=%d)\n", (int)s_pid, s_id);

    for(;;) {
        /* SQLITE_API int sqlite3_open( const char *filename, sqlite3 **ppDb); */
        s_sqlite3_check = sqlite3_open(def_test_sqlite3_database_filename, (sqlite3 **)(&s_sqlite3));
        if(s_sqlite3_check == SQLITE_OK) {
#if 0L        
            (void)fprintf(stdout, "sqlite3_open success (pid=%d, id=%d)\n", (int)s_pid, s_id);
#endif            

#if 0L
            /* SQLITE_API int sqlite3_busy_timeout(sqlite3*, int ms); */
            (void)sqlite3_busy_timeout(s_sqlite3, 5000);
#else
            /* SQLITE_API int sqlite3_busy_handler(sqlite3*, int(*)(void*,int), void*); */
            (void)sqlite3_busy_handler(s_sqlite3, test_sqlite3_busy_handler, (void *)0);
#endif
       
            (void)snprintf(
                (char *)(&s_query[0]),
                sizeof(s_query),
                "BEGIN IMMEDIATE;"
                "UPDATE %s SET %s='%s',%s='%d' WHERE %s='%s';"
                "COMMIT",
                "test_sqlite3",
                "value1", "test1",
                "value2", s_id,
                "value1", "test1"
            );
            s_sqlite3_check = test_sqlite3_query(s_sqlite3, s_pid, s_id, (const char *)(&s_query[0]));

            /* SQLITE_API int sqlite3_close(sqlite3 *); */
            s_sqlite3_check = sqlite3_close(s_sqlite3);
            if(s_sqlite3_check == SQLITE_OK) {
                (void)fprintf(stdout, "[%d:%d]", (int)s_pid, s_id);(void)fflush(stdout);
            }
            else {
                (void)fprintf(stderr, "sqlite3_close failed (pid=%d, id=%d) : error_code=%d, error_message=\"%s\"\n", (int)s_pid, s_id, s_sqlite3_check, sqlite3_errmsg(s_sqlite3));
            }
            s_sqlite3 = (sqlite3 *)0;
        }
        else {
            (void)fprintf(stderr, "sqlite3_open failed (pid=%d, id=%d) : error_code=%d\n", (int)s_pid, s_id, s_sqlite3_check);
        }

        usleep(10000);
    }

    return((void *)0);
}

static int test_sqlite3_callback(void *s_user_context, int s_argc, char **s_argv, char **s_field_name)
{
    (void)s_user_context;
    (void)s_argc;
    (void)s_argv;
    (void)s_field_name;

    return(SQLITE_OK);
}

int main(int s_argc, char **s_argv)
{
    int s_sqlite3_check;

    sqlite3 *s_sqlite3;
    char *s_error_message;
    pid_t s_pid;

    pthread_t s_pthread[ def_test_sqlite3_thread_count ];
    int s_thread_id[ def_test_sqlite3_thread_count ];
    int s_count;

    (void)s_argc;
    (void)s_argv;
    
    (void)fprintf(stdout, "initializing db... (thread_safe=%d)\n", sqlite3_threadsafe());
    /* SQLITE_API int sqlite3_open( const char *filename, sqlite3 **ppDb); */
    s_sqlite3_check = sqlite3_open(def_test_sqlite3_database_filename, (sqlite3 **)(&s_sqlite3));
    if(s_sqlite3_check != SQLITE_OK) {
        (void)fprintf(stderr, "sqlite3_open failed : error_code=%d\n", s_sqlite3_check);
        return(EXIT_FAILURE);
    }
    
    (void)fprintf(stdout, "exec #1...\n");
    /* SQLITE_API int sqlite3_exec(sqlite3*, const char *sql, int (*callback)(void*,int,char**,char**), void *, char **errmsg); */
    s_sqlite3_check = sqlite3_exec(s_sqlite3, def_test_sqlite3_drop_schema, test_sqlite3_callback, (void *)0, (char **)(&s_error_message));
    if(s_sqlite3_check != SQLITE_OK) {
        (void)fprintf(stderr, "sqlite3_exec failed : error_code=%d, error_message=\"%s\"\n", s_sqlite3_check, s_error_message);
        return(EXIT_FAILURE);
    }
    
    (void)fprintf(stdout, "exec #2...\n");
    /* SQLITE_API int sqlite3_exec(sqlite3*, const char *sql, int (*callback)(void*,int,char**,char**), void *, char **errmsg); */
    s_sqlite3_check = sqlite3_exec(s_sqlite3, def_test_sqlite3_schema, test_sqlite3_callback, (void *)0, (char **)(&s_error_message));
    if(s_sqlite3_check != SQLITE_OK) {
        (void)fprintf(stderr, "sqlite3_exec failed : error_code=%d, error_message=\"%s\"\n", s_sqlite3_check, s_error_message);
        return(EXIT_FAILURE);
    }
    
    (void)fprintf(stdout, "exec #3...\n");
    /* SQLITE_API int sqlite3_exec(sqlite3*, const char *sql, int (*callback)(void*,int,char**,char**), void *, char **errmsg); */
    s_sqlite3_check = sqlite3_exec(s_sqlite3, def_test_sqlite3_insert1, test_sqlite3_callback, (void *)0, (char **)(&s_error_message));
    if(s_sqlite3_check != SQLITE_OK) {
        (void)fprintf(stderr, "sqlite3_exec failed : error_code=%d, error_message=\"%s\"\n", s_sqlite3_check, s_error_message);
        return(EXIT_FAILURE);
    }
    
    (void)fprintf(stdout, "exec #4...\n");
    /* SQLITE_API int sqlite3_exec(sqlite3*, const char *sql, int (*callback)(void*,int,char**,char**), void *, char **errmsg); */
    s_sqlite3_check = sqlite3_exec(s_sqlite3, def_test_sqlite3_insert2, test_sqlite3_callback, (void *)0, (char **)(&s_error_message));
    if(s_sqlite3_check != SQLITE_OK) {
        (void)fprintf(stderr, "sqlite3_exec failed : error_code=%d, error_message=\"%s\"\n", s_sqlite3_check, s_error_message);
        return(EXIT_FAILURE);
    }
    
    (void)fprintf(stdout, "exec #5...\n");
    /* SQLITE_API int sqlite3_exec(sqlite3*, const char *sql, int (*callback)(void*,int,char**,char**), void *, char **errmsg); */
    s_sqlite3_check = sqlite3_exec(s_sqlite3, def_test_sqlite3_insert3, test_sqlite3_callback, (void *)0, (char **)(&s_error_message));
    if(s_sqlite3_check != SQLITE_OK) {
        (void)fprintf(stderr, "sqlite3_exec failed : error_code=%d, error_message=\"%s\"\n", s_sqlite3_check, s_error_message);
        return(EXIT_FAILURE);
    }

    (void)fprintf(stdout, "closing...\n");
    /* SQLITE_API int sqlite3_close(sqlite3 *); */
    s_sqlite3_check = sqlite3_close(s_sqlite3);
    if(s_sqlite3_check != SQLITE_OK) {
        (void)fprintf(stderr, "sqlite3_close failed : error_code=%d, error_message=\"%s\"\n", s_sqlite3_check, sqlite3_errmsg(s_sqlite3));
        return(EXIT_FAILURE);
    }
    (void)fprintf(stdout, "initialized db...\n");
    
    for(s_count = 0;s_count < def_test_sqlite3_process_count;s_count++) {
        s_pid = fork();
        if(s_pid == ((pid_t)(-1))) {
            break;
        }
        if(s_pid == ((pid_t)0)) {
            break;
        }
        (void)fprintf(stdout, "fork to (%d)...\n", (int)s_pid);
    }

    for(s_count = 0;s_count < def_test_sqlite3_thread_count;s_count++) {
        (void)fprintf(stdout, "creating thread (%d/%d)\n", s_count + 1, def_test_sqlite3_thread_count);
       
        s_thread_id[s_count] = s_count;
        if(pthread_create((pthread_t *)(&s_pthread[s_count]), (pthread_attr_t *)0, test_sqlite3_thread, (void *)(&s_thread_id[s_count])) != 0) {
            (void)fprintf(stderr, "pthread_create failed ! (id=%d)\n", s_thread_id[s_count]);
            return(EXIT_FAILURE);
        }

        (void)pthread_detach(s_pthread[s_count]);
    }

    (void)fprintf(stdout, "main loop...\n");
    (void)test_sqlite3_thread((void *)0);

    return(EXIT_SUCCESS);
}

/* vim: set expandtab: */
/* End of source */
